#!/usr/bin/env python3

"""
	CopyLeft 2019 Pascal Engélibert <tuxmain@zettascript.org>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU Affero General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Affero General Public License for more details.

	You should have received a copy of the GNU Affero General Public License
	along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import random, json, sys, math, threading, time
from concurrent.futures import ThreadPoolExecutor

CLIENT = 1
NODE = 2
lock = threading.RLock()

"""
l = [ [o, p], ... ]
 o = arbitrary value
 p = probability (0 <= p <= 1)
Returns a `o`
"""
def randlp(l:list):
	last_prob = []
	n = random.random()
	for i in range(len(l)):
		p = l[i][1]
		o = l[i][0]
		minv = 0
		for prob in last_prob:
			minv += prob
		maxv = minv + p
		last_prob.append(p)
		if n >= minv and n <= maxv:
			return o

def randdf(a, b, l):
	while True:
		n = random.randint(a, b)
		if n in l:
			continue
		return n

class Tx:
	def __init__(self, t, sender, receiver, amount):
		self.t = t
		self.sender = sender
		self.receiver = receiver
		self.amount = amount

"""
mt = number of blocks
nodes = number of nodes
clients = number of clients
txbc = tx per block per client (avg)
amounts = different possible amounts
txmin = minimum number of txs to mix
layers = number of layers in a path
"""
def simulate(mt=2016, nodes=6, clients=20, txbc=0.0035, amounts=[[100,1]], txmin=5, layers=3, expire=288, smart_clients=True, print_pending=False):
	txs = []
	pending = [[] for i in range(nodes)]
	for t in range(mt):
		
		for client in range(clients):
			if random.random() < txbc:
				left = list(range(nodes))
				random.shuffle(left)
				
				if smart_clients:
					# Remove nodes which have txmin or more pending txs
					to_remove = []
					for node in left:
						if len(pending[node]) >= txmin:
							to_remove.append(node)
					to_remove.reverse()
					for node in to_remove:
						left.pop(node)
					
					left.sort(key=lambda i: len(pending[i]), reverse=True)
					
					if print_pending:
						print([len(i) for i in pending])
					
					# Add other nodes until having enough (with the least pending txs)
					while len(left) < layers:
						mini = 0
						minp = len(pending[to_remove[0]])
						for i in range(len(to_remove)):
							p = len(pending[to_remove[i]])
							if p <= minp:
								mini = i
								minp = p
						
						to_remove.remove(mini)
						left.append(mini)
				
				path = []
				for i in range(layers):
					path.append(left.pop(0))
				path.append(random.randint(0, clients-1))
				tx = Tx(t, [CLIENT, client], [NODE if layers>0 else CLIENT, path[0]], randlp(amounts))
				pending[path[0]].append([tx, path[1:]])
				txs.append(tx)
		
		for node in range(nodes):
			expired = []
			for i in range(len(pending[node])):
				if t - pending[node][i][0].t >= expire:
					expired.append(i)
			expired.reverse()
			for i in expired:
				pending[node].pop(i)
			
			n = 0
			for tx in pending[node]:
				if tx[0].t != t:
					n += 1
			
			if n >= txmin:
				for tx in pending[node]:
					if tx[0].t == t:
						continue
					output = len(tx[1]) == 1
					ntx = Tx(t, [NODE, node], [CLIENT if output else NODE, tx[1][0]], tx[0].amount)
					if not output:
						pending[tx[1][0]].append([ntx, tx[1][1:]])
					txs.append(ntx)
				pending[node] = []
	
	return txs

def gen_graph(txs, nodes=6, clients=20):
	import igraph
	g = igraph.Graph()
	
	g.add_vertices(nodes + clients)
	
	for tx in txs:
		g.add_edges([[tx.sender[1] if tx.sender[0] == NODE else (nodes+tx.sender[1]), tx.receiver[1] if tx.receiver[0] == NODE else (nodes+tx.receiver[1])]])
	
	igraph.plot(g, "txmix.png", layout=g.layout("fr"))

def gen_tsv(txs):
	i = 0
	for tx in txs:
		print("{}\t{}\t{}{}\t{}{}".format(i, tx.t, "C" if tx.sender[0] == CLIENT else "N", tx.sender[1], "C" if tx.receiver[0] == CLIENT else "N", tx.receiver[1]))
		i += 1

def last(l):
	return l[len(l)-1]

def retro_tx(txs, txid, max_depth=3):
	paths = []
	tx = txs[txid]
	i = 0
	while txs[i].t < tx.t:
		if txs[i].receiver == tx.sender and txs[i].amount == tx.amount:
			if txs[i].sender[0] == CLIENT:
				paths.append(txs[i].sender[1])
			elif max_depth > 0:
				paths.append(retro_tx(txs, i, max_depth-1))
		i += 1
	
	return paths

def filter_layers(retro, layers=3, recursive=False):
	if not recursive:
		layers -= 1
	result = []
	for i in retro:
		if layers == 0:
			if isinstance(i, int):
				result.append(i)
		else:
			if isinstance(i, list):
				p = filter_layers(i, layers-1, recursive=True)
				if len(p) > 0:
					result.append(p)
	return result

def count_hits(retro):
	result = {}
	for i in retro:
		if isinstance(i, int):
			if i in result:
				result[i] += 1
			else:
				result[i] = 1
		else:
			r = count_hits(i)
			for j in r:
				if j in result:
					result[j] += r[j]
				else:
					result[j] = r[j]
	return result

def count_proba(hits):
	s = sum([hits[i] for i in hits])
	return {i:hits[i]/s for i in hits}

def json_to_txs(data):
	return [Tx(**tx) for tx in json.loads(data)]

def txs_to_json(txs):
	return json.dumps([{"t": tx.t, "sender": tx.sender, "receiver": tx.receiver, "amount": tx.amount} for tx in txs])

def stddev(l):
	n = len(l)
	avg = sum(l)/n
	return math.sqrt(sum([(avg-i)**2 for i in l])/n)

# Extracts arg value from argv
def getargv(arg, default="") -> str:
	if arg in sys.argv and len(sys.argv) > sys.argv.index(arg)+1:
		return sys.argv[sys.argv.index(arg)+1]
	else:
		return default

def analyze(txs, txid, layers, results):
	r = retro_tx(txs, txid, layers)
	fl = filter_layers(r, layers)
	hits = count_hits(fl)
	proba = count_proba(hits)
	if len(proba) > 0:
		res = [txid, stddev([proba[i] for i in proba]), len(proba)]
		with lock:
			results.append(res)
			sys.stderr.write(str(len(results))+"\r")

if __name__ == "__main__":
	if "--help" in sys.argv:
		print("""ĞMixer simulator & analyzer

Output:
 List of output txs in TSV format
 (<block_id> <std_deviation> <nb>)

Options:
 -l <int>    layers (default: 3)
 -t <int>    blocks (default: 2016)
 -n <int>    nodes (default: 6)
 -c <int>    clients (default: 20)
 -x <float>  avg txs per client per block (default: 0.0035)
 -m <int>    min txs for mixing (default: 5)
 -e <int>    expiration time of tx, in blocks (default: 288)
 -s <0|1>    enable smart clients (default: 1)
             If 1, clients choose nodes by an optimized way.
 -h <int>    number of threads (default: 4)
 --print_pending  print pending txs during generation
 --print_sets     print settings
 --csv       print in CSV format instead of TSV
 tsv         print all txs in TSV format
""")
		exit()
	
	layers = int(getargv("-l", 3))
	
	settings = {
		"mt" : int(getargv("-t", 2016)),
		"nodes" : int(getargv("-n", 6)),
		"clients" : int(getargv("-c", 20)),
		"txbc" : float(getargv("-x", 0.0035)),
		"amounts" : json.loads(getargv("-a", "[[100,1]]")),
		"txmin" : int(getargv("-m", 5)),
		"layers" : layers,
		"expire" : int(getargv("-e", 288)),
		"smart_clients" : bool(int(getargv("-s", 1))),
		"print_pending" : "--print_pending" in sys.argv
	}
	
	txs = simulate(**settings)
	
	if "--print_sets" in sys.argv:
		print(json.dumps(settings))
	
	if "tsv" in sys.argv:
		gen_tsv(txs)
		exit()
	
	executor = ThreadPoolExecutor(int(getargv("-h", 4)))
	results = []
	nth = 0
	with lock:
		for txid in range(len(txs)):
			if txs[txid].receiver[0] != CLIENT:
				continue
			
			f=executor.submit(analyze, txs, txid, layers, results)
			nth += 1
		
		sys.stderr.write(str(nth)+"\n")
	
	executor.shutdown()
	
	if "--csv":
		for i in results:
			print("{},{},{}".format(txs[i[0]].t, round(i[1], 6), i[2]))
	else:
		for i in results:
			print(txs[i[0]].t, "\t", round(i[1], 6), "\t", i[2])
